import sys
import os
from glob import glob
import webbrowser
from flask import Flask, render_template, url_for, request
import time
import threading
from pathlib import Path

sys.dont_write_bytecode = True
from ffmpeg4discord import arguments
from ffmpeg4discord.twopass import TwoPass, seconds_to_timestamp


def twopass_loop(twopass: TwoPass, target_filesize: float, approx: bool = False) -> None:
    while twopass.run() >= target_filesize and not approx:
        print(
            f"\nThe output file size ({round(twopass.output_filesize, 2)}MB) is still above the target of {target_filesize}MB.\nRestarting...\n"
        )
        os.remove(twopass.output_filename)

        # adjust the class's target file size to set a lower bitrate for the next run
        twopass.target_filesize -= 0.2

    # clean up
    for file in glob("ffmpeg2pass*"):
        os.remove(file)

    twopass.message = f"Your compressed video file ({round(twopass.output_filesize, 2)}MB) is located at {Path(twopass.output_filename).resolve()}"


def open_browser(port: int) -> None:
    time.sleep(0.5)
    webbrowser.open(f"http://localhost:{port}")


def main() -> None:
    # get args from the command line
    args = arguments.get_args()
    web = args.pop("web")
    approx = bool(args.pop("approx"))

    if web:
        port = args.pop("port")

    path = Path(args["filename"]).resolve()
    args["filename"] = path

    # instantiate the TwoPass class
    twopass = TwoPass(**args)

    if web:
        app = Flask(__name__, static_folder=path.parent)

        @app.route("/")
        def index():
            return render_template(
                "web.html",
                file_url=url_for("static", filename=path.name),
                twopass=twopass,
                alert_hidden=True,
                approx=approx,
            )

        @app.route("/encode", methods=["POST"])
        def form_twopass():
            # generate new times from the selection
            ss = int(request.form.get("startTime"))
            to = int(request.form.get("endTime"))
            twopass.length = to - ss
            twopass.times = {"ss": seconds_to_timestamp(ss), "to": seconds_to_timestamp(to)}
            target_filesize = float(request.form.get("target_filesize"))

            # update TwoPass from web form
            twopass.resolution = request.form.get("resolution")
            twopass.target_filesize = target_filesize
            twopass.crop = request.form.get("crop")
            twopass.output = Path(request.form.get("output"))
            twopass.codec = request.form.get("codec")
            twopass.framerate = int(framerate) if (framerate := request.form.get("framerate")) else None
            twopass.audio_br = (
                float(audio_br) * 1000 if (audio_br := request.form.get("audio_br")) else twopass.audio_br
            )

            # to loop or not to loop
            approx = bool(request.form.getlist("approx"))

            twopass_loop(twopass=twopass, target_filesize=target_filesize, approx=approx)

            return render_template(
                "web.html",
                file_url=url_for("static", filename=path.name),
                twopass=twopass,
                approx=approx,
            )

        threading.Thread(target=open_browser, args=[port], name="Open Browser").start()
        app.run("0.0.0.0", port=port)
    else:
        twopass_loop(twopass=twopass, target_filesize=twopass.target_filesize, approx=approx)
        print(twopass.message)


if __name__ == "__main__":
    main()



import sys
import os
from urllib.request import urlretrieve
import zipfile
from pathlib import Path
import shutil
import logging
import platform

if platform.system() != "Windows":
    logging.warning(
        """
        THIS SCRIPT IS INTENDED FOR WINDOWS USERS

        If you are running this script on macOS, consider installing ffmpeg with
        Homebrew: https://formulae.brew.sh/formula/ffmpeg

        If you are running this script on some Linux, learn more about your options
        to install ffmpeg on its website: https://ffmpeg.org/download.html#build-linux
        """
    )

    sys.exit()

logging.getLogger().setLevel(logging.INFO)

path = os.environ.get("PATH", "")
python_install = Path(sys.executable)

if "WindowsApps" in str(python_install):
    print(
        "Microsoft Store Python installation detected. If you run into problems, consider using the https://python.org installation instead!"
    )
    target = Path(sys.executable).parent.parent
else:
    print("python.org Python installation detected.")
    target = Path(sys.executable).parent / "Scripts"

target.mkdir(parents=True, exist_ok=True)

if str(target) not in path:
    logging.warning(
        """
        The directory we are installing ffmpeg into is not in your 
        system's PATH. Windows will not be able to find ffmpeg when 
        trying to run ffmpeg4discord. Please ensure that you installed 
        Python with the "Add Python to PATH" option selected.
        
        More information: https://docs.python.org/3/using/windows.html#finding-the-python-executable
        """
    )


def copy_directory_contents(source: Path, target: Path) -> None:
    source_path = source
    target_path = target

    # Iterate over files in the source directory
    for item in source_path.iterdir():
        # Construct target item path
        target_item = target_path / item.name

        # If the item is a file, copy it to the target directory
        if item.is_file():
            shutil.copy2(item, target_item)
            print(f"Copied file: {item} -> {target_item}")


def download_with_progress(url: str, save_path: str) -> None:
    def report(block_num, block_size, total_size):
        downloaded = block_num * block_size
        downloaded_mb = downloaded / (1024 * 1024)
        total_size_mb = total_size / (1024 * 1024)
        progress = min(downloaded / total_size, 1.0)
        percent = round(progress * 100, 2)
        print(f"\rDownloaded {downloaded_mb:.2f}/{total_size_mb:.2f} MB ({percent}%)", end="")

    save_dir = Path(save_path).parent
    save_dir.mkdir(parents=True, exist_ok=True)

    urlretrieve(url, save_path, reporthook=report)
    print("\nDownload complete!")


def install() -> None:
    print("Downloading ffmpeg to ffmpeg.zip...")
    download_with_progress("https://www.gyan.dev/ffmpeg/builds/ffmpeg-release-essentials.zip", "ffmpeg/ffmpeg.zip")

    print("\nUnzipping ffmpeg.zip...")
    with zipfile.ZipFile("ffmpeg/ffmpeg.zip", "r") as zip_ref:
        zip_ref.extractall("ffmpeg/")
        source = Path(f"ffmpeg/{zip_ref.namelist()[0]}/bin").resolve()
    print("Done!\n")

    copy_directory_contents(source=source, target=target)
    shutil.rmtree("ffmpeg/")

    print("\nffmpeg installation complete.")
    print("\nOpen a new Terminal window and type ffmpeg and hit enter to verify that the installation succeeded.")


if __name__ == "__main__":
    install()



import unittest
from unittest.mock import patch, MagicMock
from ffmpeg4discord.twopass import TwoPass
from pathlib import Path


class TestTwoPass(unittest.TestCase):
    def setUp(self):
        self.filename = Path("000100.mp4")
        self.target_filesize = 50.0  # MB

    @patch("ffmpeg4discord.twopass.ffmpeg.probe")
    def test_init(self, mock_probe: MagicMock):
        # Set up mock values for the probe
        mock_probe.return_value = {
            "streams": [
                {"index": 0, "codec_type": "video", "width": 1280, "height": 720, "r_frame_rate": "60/1"},
                {"index": 1, "codec_type": "audio", "bit_rate": "128000"},
            ],
            "format": {"duration": "3600"},
        }

        # Create TwoPass instance
        twopass = TwoPass(self.filename, self.target_filesize)

        # Check attributes
        self.assertEqual(twopass.filename, self.filename)
        self.assertEqual(twopass.target_filesize, self.target_filesize)
        self.assertEqual(twopass.duration, 3600)
        self.assertEqual(twopass.audio_br, 128000)

    @patch("ffmpeg4discord.twopass.ffmpeg.probe")
    def test_time_from_file_name(self, mock_probe: MagicMock):
        # Set up mock values for the probe
        mock_probe.return_value = {
            "streams": [
                {"index": 0, "codec_type": "video", "width": 1280, "height": 720, "r_frame_rate": "60/1"},
                {"index": 1, "codec_type": "audio", "bit_rate": "128000"},
            ],
            "format": {"duration": "3600"},
        }
        # mock_seconds_from_ts_string.return_value = 60

        # Create TwoPass instance
        twopass = TwoPass(self.filename, self.target_filesize)

        # Call time_from_file_name
        twopass.time_from_file_name()

        # Check attributes
        self.assertEqual(twopass.from_seconds, 60)
        self.assertEqual(twopass.times, {"ss": "00:01:00", "to": "01:00:00"})

    @patch("ffmpeg4discord.twopass.ffmpeg.probe")
    def test_create_bitrate_dict(self, mock_probe: MagicMock):
        # Set up mock values for the probe
        mock_probe.return_value = {
            "streams": [
                {"index": 0, "codec_type": "video", "width": 1280, "height": 720, "r_frame_rate": "60/1"},
                {"index": 1, "codec_type": "audio", "bit_rate": "128000"},
            ],
            "format": {"duration": "120"},
        }

        # Create TwoPass instance
        twopass = TwoPass(self.filename, self.target_filesize)

        # Call create_bitrate_dict
        twopass.create_bitrate_dict()

        # Check bitrate_dict attribute
        self.assertEqual(twopass.bitrate_dict["b:v"], 3285000)
        self.assertEqual(twopass.bitrate_dict["minrate"], 1642500)
        self.assertEqual(twopass.bitrate_dict["maxrate"], 4763250)
        self.assertEqual(twopass.bitrate_dict["bufsize"], 6570000)

    @patch("ffmpeg4discord.twopass.os.path.getsize")
    @patch("ffmpeg4discord.twopass.ffmpeg.probe")
    @patch("ffmpeg4discord.twopass.ffmpeg.output")
    def test_run(self, mock_output: MagicMock, mock_probe: MagicMock, mock_os_path_getsize: MagicMock):
        # Set up mock values for the probe
        mock_probe.return_value = {
            "streams": [
                {"index": 0, "codec_type": "video", "width": 1280, "height": 720, "r_frame_rate": "60/1"},
                {"index": 1, "codec_type": "audio", "bit_rate": "128000"},
            ],
            "format": {"duration": "3600"},
        }

        # Create TwoPass instance
        twopass = TwoPass(self.filename, self.target_filesize)

        # Mock output.run() to return some values
        mock_run = MagicMock()
        mock_run.run.return_value = ("a", "b")
        mock_output.return_value = mock_run
        mock_output.return_value.global_args.return_value = mock_run

        # Fake a file size
        mock_os_path_getsize.return_value = 52428799

        # Call run method
        result = twopass.run()

        # Check output_filesize attribute and return value
        self.assertLess(twopass.output_filesize, 50)  # Mocking output file size
        self.assertLess(result, 50)


if __name__ == "__main__":
    unittest.main()



import json
import logging
import socket

from argparse import ArgumentParser, Namespace, BooleanOptionalAction
from pathlib import Path
from random import randint


def is_port_in_use(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


def get_args() -> Namespace:
    parser = ArgumentParser(
        prog="ffmpeg4discord",
        description="This script takes a video file and compresses it to a target file size.",
        epilog="For more help: https://github.com/zfleeman/ffmpeg4discord",
    )
    parser.add_argument("filename", help="The file path of the file that you wish to compress.")
    parser.add_argument(
        "-o",
        "--output",
        default="",
        help="The desired output directory where the file will land.",
    )
    parser.add_argument(
        "--filename-times",
        action=BooleanOptionalAction,
        help="Generate From/To timestamps from the clip's file name.",
    )
    parser.add_argument(
        "--approx",
        action=BooleanOptionalAction,
        help="Approximate file size. The job will not loop to output the file under the target size. It will get close enough to the target on the first run.",
    )
    parser.add_argument("--from", help="Start clipping at this timestamp, e.g. 00:00:10")
    parser.add_argument("--to", help="Stop clipping at this timestamp, e.g. 00:00:20")
    parser.add_argument(
        "-s",
        "--target-filesize",
        default=10,
        type=float,
        help="The output file size in MB.",
    )
    parser.add_argument("-a", "--audio-br", type=float, default=96, help="Audio bitrate in kbps.")
    parser.add_argument(
        "-c", "--codec", type=str, default="libx264", choices=["libx264", "libvpx-vp9"], help="Video codec."
    )
    parser.add_argument(
        "--vp9-opts",
        type=str,
        default=None,
        help="""JSON string to configure row-mt, deadline, and cpu-used options for VP9 encoding. (e.g., --vp9-opts \'{"row-mt": 1, "deadline": "good", "cpu-used": 2}\')')""",
    )

    # video filters
    parser.add_argument("-x", "--crop", default="", help="Cropping dimensions. Example: 255x0x1410x1080")
    parser.add_argument("-r", "--resolution", default="", help="The output resolution of your final video.")
    parser.add_argument("-f", "--framerate", type=int, help="The desired output frames per second.")

    # configuraiton json file
    parser.add_argument("--config", help="JSON file containing the run's configuration")

    # web
    parser.add_argument("--web", action=BooleanOptionalAction, help="Launch ffmpeg4discord's Web UI in your browser.")
    parser.add_argument("-p", "--port", type=int, help="Local port for the Flask application.")

    args = vars(parser.parse_args())

    # fill in from the config JSON
    if args["config"]:
        file_path = Path(args["config"]).resolve()
        with open(file_path) as f:
            config = json.load(f)

        for k, v in config.items():
            if not args[k]:
                args[k] = v

    # do some work regarding the port
    if args["web"]:
        port = args.pop("port")
        port = port if port else randint(5000, 6000)

        while True:
            if not is_port_in_use(port):
                break
            logging.warning(f"Port {port} is already in use. Retrying with a new port.")
            port = randint(5000, 6000)

        args["port"] = port
    else:
        del args["port"]

    args["times"] = {}

    # adjust times
    if args["from"]:
        args["times"]["from"] = args["from"]

    if args["to"]:
        args["times"]["to"] = args["to"]

    del args["from"]
    del args["to"]

    # work with vp9 options
    if args["vp9_opts"]:
        logging.info(f"Received VP9 options: {args['vp9_opts']}")
        try:
            args["vp9_opts"] = json.loads(args["vp9_opts"])
            if not isinstance(args["vp9_opts"], dict):
                logging.error("The `vp9-opts` input must be a dictionary. Using default parameters.")
                args["vp9_opts"] = None
        except json.JSONDecodeError:
            logging.error(
                """Invalid JSON format. Format your input string like this: \'{"row-mt": 1, "deadline": "good", "cpu-used": 2}\'. Using default parameters."""
            )
            args["vp9_opts"] = None

    return args



import json
import logging
import math
import os

from datetime import datetime
from pathlib import Path
from typing import Optional

import ffmpeg

logging.getLogger().setLevel(logging.INFO)


class TwoPass:
    """
    Encodes and resizes video files using ffmpeg's two-pass encoding to meet a specified target file size.

    The TwoPass class leverages the ffmpeg-python library to compress video files while allowing control
    over attributes such as codec, resolution, and target file size. It performs two-pass encoding
    to achieve optimal video quality at the desired size.

    Two-Pass encoding documentation: https://trac.ffmpeg.org/wiki/Encode/H.264#twopass

    Attributes:
        filename (Path): Path to the input video file that needs compression.
        target_filesize (float): Desired target filesize in megabytes (MB).
        output (str): Output file path or directory where the compressed video will be saved.
        times (dict): Dictionary with keys "from" and "to" specifying timestamps (in seconds) for encoding a segment.
        audio_br (float): Audio bitrate in kilobits per second (kbps), if specified. Defaults to automatic calculation.
        codec (str): Video codec to use for compression, e.g., 'libx264' (default).
        crop (str): Crop settings (if any) for the video.
        resolution (str): Target resolution for the output video.
        config (str): Path to an optional configuration file for advanced ffmpeg settings.
        filename_times (bool): Flag to include timestamps in the output filename.
    """

    def __init__(
        self,
        filename: Path,
        target_filesize: float,
        output: str = "",
        times: Optional[dict] = None,
        audio_br: Optional[float] = None,
        codec: str = "libx264",
        crop: str = "",
        resolution: str = "",
        config: Optional[str] = None,
        filename_times: bool = False,
        framerate: Optional[int] = None,
        vp9_opts: Optional[dict] = None,
    ) -> None:

        self.target_filesize = target_filesize
        self.crop = crop
        self.resolution = resolution
        self.times = times or {}
        self.audio_br = audio_br
        self.codec = codec
        self.framerate = framerate
        self.output = output
        self.vp9_opts = vp9_opts or {}

        self.filename = filename
        self.fname = filename.name
        self.split_fname = self.fname.split(".")

        # create a Path from the output string
        self.output = Path(self.output).resolve()

        self.probe = ffmpeg.probe(filename=filename)
        self.duration = math.floor(float(self.probe["format"]["duration"]))

        if len(self.probe["streams"]) > 2:
            logging.warning(
                "This media file has more than two streams, which could cause errors during the encoding job."
            )

        # Extract some information from the probe.
        for stream in self.probe["streams"]:
            ix = stream["index"]
            codec_type = stream["codec_type"]
            if codec_type == "video":
                width = self.probe["streams"][ix]["width"]
                height = self.probe["streams"][ix]["height"]
                self.ratio = width / height

                # Get the framerate for later comparisons
                framerate_ratio: str = self.probe["streams"][ix].get("r_frame_rate")
                self.init_framerate = round(int(framerate_ratio.split("/")[0]) / int(framerate_ratio.split("/")[1]))

            elif codec_type == "audio":
                audio_stream = ix

        if not self.audio_br:
            self.audio_br = float(self.probe["streams"][audio_stream]["bit_rate"])
        else:
            self.audio_br = self.audio_br * 1000

        # times are supplied by the file's name
        if filename_times:
            self.time_from_file_name()

        # times are provided by the flags or config file
        elif self.times:
            if self.times.get("from"):
                self.times["ss"] = self.times["from"] or "00:00:00"
                del self.times["from"]
            else:
                self.times["ss"] = "00:00:00"

            self.from_seconds = seconds_from_ts_string(self.times["ss"])

            if self.times.get("to"):
                self.to_seconds = seconds_from_ts_string(self.times["to"])
                self.length = self.to_seconds - self.from_seconds
            else:
                self.length = self.duration - self.from_seconds
                self.times["to"] = seconds_to_timestamp(self.duration)

        # no trimming times were provided
        else:
            self.times = {"ss": "00:00:00", "to": seconds_to_timestamp(self.duration)}
            self.length = self.duration

        if self.length <= 0:
            raise Exception(
                f"""
                Time Paradox?

                Something is wrong with your clipping times. Use this
                information to further diagnose the problem:

                - Your video is {self.duration / 60} minutes long
                - Your clipping times are {self.times}
                """
            )

    def generate_params(self, codec: str) -> dict:
        """
        Create params for the ffmpeg.output() function
        :param codec: ffmpeg video codec to use during encoding
        :return: dictionary containing parameters for ffmpeg's first and second pass.
        """

        params = {
            "pass1": {
                "pass": 1,
                "f": "null",
                "vsync": "cfr",  # not sure if this is unique to x264 or not
                "c:v": codec,
            },
            "pass2": {"pass": 2, "b:a": self.audio_br, "c:v": codec},
        }

        # assign the output framerate
        if self.framerate:
            if self.framerate < self.init_framerate:
                params["pass1"]["r"] = self.framerate
                params["pass2"]["r"] = self.framerate
            else:
                logging.warning(
                    f"Your output framerate ({self.framerate}) is more than the original framerate ({self.init_framerate}). Keeping the original framerate..."
                )

        if codec == "libx264":
            params["pass2"]["c:a"] = "aac"
        elif codec == "libvpx-vp9":
            row_mt = self.vp9_opts.get("row-mt", 1)
            cpu_used = self.vp9_opts.get("cpu-used", 2)
            deadline = self.vp9_opts.get("deadline", "good")

            params["pass1"]["row-mt"] = row_mt
            params["pass2"]["row-mt"] = row_mt
            params["pass2"]["cpu-used"] = cpu_used
            params["pass2"]["deadline"] = deadline
            params["pass2"]["c:a"] = "libopus"

        params["pass1"].update(**self.bitrate_dict)
        params["pass2"].update(**self.bitrate_dict)

        return params

    def create_bitrate_dict(self) -> None:
        """
        Perform the calculation specified in ffmpeg's documentation that generates
        the video bitrates needed to achieve the target file size
        """
        br = math.floor((self.target_filesize * 8192) / self.length - (self.audio_br / 1000)) * 1000
        self.bitrate_dict = {
            "b:v": br,
            "minrate": br * 0.5,
            "maxrate": br * 1.45,
            "bufsize": br * 2,
        }

    def time_from_file_name(self):
        """
        Extracts start (`-ss`) and optional end (`-to`) timestamps from the file name and updates the instance's
        time-related attributes. Assumes the file name is in one of two formats:
        - 'HHMMSS' (start time only)
        - 'HHMMSS-HHMMSS' (start and end times)

        If the end time is not provided, the function defaults the end time to the full video duration.
        """
        fname = self.fname
        times = {}

        try:
            # Parse start time from the first six characters
            start_time_str = f"{fname[0:2]}:{fname[2:4]}:{fname[4:6]}"
            self.from_seconds = seconds_from_ts_string(start_time_str)
            times["ss"] = start_time_str

            # Check if there is an additional six digits for end time
            if len(fname) > 6 and fname[7:13].isdigit():
                end_time_str = f"{fname[7:9]}:{fname[9:11]}:{fname[11:13]}"
                self.to_seconds = seconds_from_ts_string(end_time_str)
                times["to"] = end_time_str
                self.length = self.to_seconds - self.from_seconds
            else:
                # Default to the full duration if end time is not provided
                times["to"] = seconds_to_timestamp(self.duration)
                self.length = self.duration - self.from_seconds
                self.to_seconds = self.duration

        except ValueError:
            # Handle any issues with timestamp parsing by defaulting to full duration
            self.length = self.duration
            self.times = {"ss": "00:00:00", "to": seconds_to_timestamp(self.duration)}
            logging.warning("Warning: Invalid time format in filename. Defaulting to full duration.")

        # Update instance attributes with calculated times
        self.times = times

    def apply_video_filters(self, video):
        """
        Function to apply the crop and resolution parameters to a video object
        :param video: the ffmpeg video object from the Class's input video file
        :return: the video object after it has been cropped or resized
        """

        if self.crop:
            crop = self.crop.split("x")
            video = video.crop(x=crop[0], y=crop[1], width=crop[2], height=crop[3])
            self.ratio = int(crop[2]) / int(crop[3])

        if self.resolution:
            video = video.filter("scale", self.resolution)
            x = int(self.resolution.split("x")[0])
            y = int(self.resolution.split("x")[1])
            outputratio = x / y

            if self.ratio != outputratio:
                logging.warning(
                    """
                    Your output resolution's aspect ratio does not match the
                    input resolution's or your croped resolution's aspect ratio.
                    """
                )

        return video

    def run(self) -> float:
        """
        Perform the CPU-intensive encoding job
        :return: the output file's size
        """

        ext: str = ".webm" if self.codec == "libvpx-vp9" else ".mp4"
        if self.output.is_dir():
            self.output_filename = str(
                self.output
                / (
                    "small_"
                    + self.filename.stem.replace(" ", "_")
                    + datetime.strftime(datetime.now(), f"_%Y%m%d%H%M%S{ext}")
                )
            )
        else:
            if ext != self.output.suffix:
                logging.warning(
                    f"You specified {self.codec}, but your output file name ends with {self.output.suffix}. I've corrected this."
                )

                # correct the file suffix
                if self.codec == "libvpx-vp9":
                    self.output = self.output.with_suffix(".webm")
                else:
                    self.output = self.output.with_suffix(".mp4")

            self.output_filename = str(self.output)

        # generate run parameters
        self.create_bitrate_dict()
        params = self.generate_params(codec=self.codec)

        # separate streams from ffinput
        ffinput = ffmpeg.input(self.filename, **self.times)
        video = self.apply_video_filters(ffinput.video)
        audio = ffinput.audio

        # First Pass
        ffOutput = ffmpeg.output(video, "pipe:", **params["pass1"])
        ffOutput = ffOutput.global_args("-loglevel", "quiet", "-stats")
        print("Performing first pass")
        std_out, std_err = ffOutput.run(capture_stdout=True)

        # Second Pass
        ffOutput = ffmpeg.output(video, audio, self.output_filename, **params["pass2"])
        ffOutput = ffOutput.global_args("-loglevel", "quiet", "-stats")
        print("\nPerforming second pass")
        ffOutput.run(overwrite_output=True)

        # save the output file size and return it
        self.output_filesize = os.path.getsize(self.output_filename) * 0.00000095367432

        return self.output_filesize


def seconds_from_ts_string(ts_string: str):
    return int(ts_string[0:2]) * 60 * 60 + int(ts_string[3:5]) * 60 + int(ts_string[6:8])


def seconds_to_timestamp(seconds: int) -> str:
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    # Use f-strings to format the timestamp
    timestamp = f"{hours:02d}:{minutes:02d}:{seconds:02d}"

    return timestamp
